/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "dac_adc_ovfl.hh"
#include "jsonStack.hh"
#include "FrWriter.hh"
#include "DVecType.hh"
#include "frame_name.hh"

using namespace std;

//======================================  Generate main program
EXECDAT(adc_dac_ovfl)

//======================================  Constructor
adc_dac_ovfl::adc_dac_ovfl(int argc, const char *argv[])
 : DatEnv(argc, argv), startTime(0), endTime(0), currentTime(0), stride(0),
   framesout(0)
{
   configure(cin, "cin");
   for (auto iter=adc_modvec.begin(); iter != adc_modvec.end(); iter++) {
      iter->setup(getDacc());
   }
   getDacc().setStride(stride);
   if (startTime != Time(0)) getDacc().seek(startTime);
}

//======================================  Destructor
adc_dac_ovfl::~adc_dac_ovfl(void) {
   Interval dt = stride;
   Time tStart(0);
   while (dt != Interval(0)) {
      for (auto iter=adc_modvec.begin(); iter != adc_modvec.end(); iter++) {
	 Interval l = iter->get_record().getInterval();
	 if (l < dt) {
	    dt = l;
	    tStart = iter->get_record().getStartTime();
	 }
      }
      if (dt != Interval(0)) write_data(tStart, dt);
   }
   printResult(cout);
}

//======================================  Parse configuration.
bool 
adc_dac_ovfl::configure(std::istream& file, const std::string& docname) {
   jsonStack js;
   js.parse(file, docname);

   js.fetch_data("ifo", ifo);

   js.fetch_data("name", name);
   
   int t=0;
   if (js.fetch_data("start-gps", t))  startTime = Time(t);
   if (js.fetch_data("end-gps",   t))  endTime   = Time(t);

   double dT;
   if (js.fetch_data("stride", dT)) {
      stride = dT;
   } else if (!startTime || !endTime) {
      stride = 64;
   } else {
      stride = endTime - startTime;
   }
  
   js.fetch_data("frame-template", file_path);

   //------------------------------------  Get input frame paths
   if (js.push_member("input-path")) {
      jsonStack::autopop pop1(js);
      string path;
      while (js.iterate()) {
	 jsonStack::autopop pop2(js);
	 if (js.fetch_value(path)) getDacc().addPath(path);
      }
   }
  
   //------------------------------------  Read in module configuration
   if (js.push_member("modules")) {
      jsonStack::autopop pop1(js);
      int last_dcu=0, last_unit=0;
      while (js.iterate()) {
	 jsonStack::autopop pop2(js);
	 int dcu=0, unit=0;
	 string mod_ifo = ifo;
	 js.fetch_data("ifo", mod_ifo);
	 if (!js.fetch_data("dcu-id", dcu)) dcu = last_dcu;
	 if (!dcu) throw runtime_error("dcu-id not specified");
	 if (dcu == last_dcu) unit = last_unit + 1;
	 js.fetch_data("unit", unit);
	 adc_modvec.push_back(adcMod(mod_ifo, dcu, unit));
	 adcMod& module(adc_modvec.back());
	 string scratch = "%i:FEC-%d_ADC_OVERFLOW_ACC_%u_%c";
	 js.fetch_data("channel-template", scratch);
	 module.set_channel_template(scratch);
	 scratch = "%i:FEC-%d_ADC_SIGNAL_%u_%c";
	 js.fetch_data("signal-template", scratch);
	 module.set_signal_template(scratch);
	 int mask = ~0;
	 js.fetch_data("active-mask", mask);
	 module.activeMask(mask);
	 js.fetch_data("select-mask", mask);
	 module.selectMask(mask);
	 last_unit = unit;
	 last_dcu  = dcu;
      }
   }  
}

//======================================  Print report
void 
adc_dac_ovfl::printResult(std::ostream& out) const {
   vector<string> names;
   adcMod::mask_type combined = 0;
   for (auto iter=adc_modvec.begin(); iter != adc_modvec.end(); iter++) {
      adcMod::mask_type alarms = iter->alarmBits();
      combined |= alarms;
      for (int i=0; alarms; i++) {
	 adcMod::mask_type m = 1 << i;
	 if (!(alarms & m)) continue;
	 names.push_back(iter->signal_name(i));
	 alarms  &= ~m;
      }
   }
   string result = combined ? "Alarm" : "OK";

   out << "{" << endl;
   out << "   \"name\" : \"" << name << "\"" << endl;
   out << "   \"result\" : \"" << result << "\"" << endl;
   if (combined) {
      out << "   \"overflows\" : [" << endl;
      for (size_t i=0; i < names.size(); i++) {
	 out << "   \"" << names[i] << "\"";
	 if (i+1 < names.size()) out << ",";
	 out << endl;
      }
      out << "   ]" << endl;
   }
   out << "}" << endl;
}

//======================================  process a stride
void 
adc_dac_ovfl::ProcessData(void) {
   if (!currentTime || currentTime != getDacc().getFillTime()) {
      reset();
      currentTime =  getDacc().getFillTime();
   }
   Time tEnd = getDacc().getCurrentTime();
   
   for (auto iter=adc_modvec.begin(); iter != adc_modvec.end(); iter++) {
      iter->process(getDacc());
   }

   write_data(currentTime, tEnd - currentTime);
   
   currentTime = tEnd;
   if (endTime != Time(0) && currentTime >= endTime) finish();
}


//======================================  Reset at the start of a segment
void 
adc_dac_ovfl::reset(void) {
   currentTime = Time(0);
   for (auto iter=adc_modvec.begin(); iter != adc_modvec.end(); iter++) {
      iter->reset();
   }
}

void
adc_dac_ovfl::write_data(const Time& t0, Interval dT) {
   if (file_path.empty()) return;
   
   FrWriter  writer;
   writer.buildFrame(t0, dT);
   
   for (auto iter=adc_modvec.begin(); iter != adc_modvec.end(); iter++) {
      iter->write(writer, t0, dT);
   }

   //----------------------------------  Open a writer to a frame file
   frame_name path;
   path.split(file_path);
   string outdir(path.dir_name(t0));
   if (outdir.substr(0,8) != "/online/") {
      path.make_dir(outdir);
      string file_path = path.file_path(t0, dT);
      if (Debug()) cout << "  opening file: " << file_path << endl;
      if (writer.open(file_path.c_str(), true)) {
	 writer.erase();
	 cerr << "DMTGen: Failed to open frame " << file_path
	      << " for output." << endl;
	 return;
      }
   } 

   //----------------------------------  Open a writer to shared memory.
   else {
      if (writer.open(outdir.c_str(), true)) {
	 writer.erase();
	 ostringstream msg;
	 msg << "Failed to open partition " << outdir << " for output.";
	 cerr << "DMTGen: " << msg.str() << endl;
	 throw runtime_error(msg.str());
      }
   }
   
   //----------------------------------  Write the frame
   try {
      if (writer.writeFrame()) {
	 cerr << "Error writing frame with GPS " << t0.getS()
	      << " to output Directory " << outdir << endl;
      } else {
	 framesout++;
      }
   } catch (std::exception& e) {
      cerr << "Caught exception in writeFrame: " << e.what() << endl;
   }
   
   //----------------------------------  Clean up
   writer.close();
}

//======================================  Adc Module constructor
adcMod::adcMod(const std::string& ifo, int dcu, int unit)
   : ifoid(ifo), dcuid(dcu), unitid(unit), active_mask(~0), select_mask(0),
     chan0_offset(0), debug(0)
{
}

//======================================  Adc Module destructor
adcMod::~adcMod(void) {
}

//======================================
void
adcMod::activeMask(mask_type mask) {
   active_mask = mask;
}

//======================================  Process a data stride
void
adcMod::process(DaccAPI& in) {
   bool begin_seg = false;
   if (!tCurrent || in.getFillTime() != tCurrent) {
      begin_seg = true;
      tCurrent = in.getFillTime();
   } 
   DVectU dvu;
   Time tStart(0);
   Interval dT(0);
   mask_type active_left = active_mask;
   for (size_t i=0; active_left; i++) {
      mask_type m = 1 << i;
      if (!(m & active_mask)) continue;
      auto ts = in.refData(channels[i]);
      if (!ts || ts->empty()) {
	 cerr << "Channel: " << channels[i] << " not found. Deactivated."
	      << endl;
	 active_mask &= ~m;
	 active_left &= ~m;
	 continue;
      }
      size_t N=ts->getNSample();
      if (!tStart) {
	 tStart = ts->getStartTime();
	 dvu.Extend(N);
	 dT = ts->getTStep();
      }
      const DVectI& tsv = dynamic_cast<const DVectI&>(*(ts->refDVect()));
      int count = begin_seg ? tsv[0] : prev[i];
      for (size_t j=0; j<N; j++) {
	 if (tsv[j] > count) {
	    dvu[j] |= m;
	    alarm_flags |= (m & select_mask);
	 }
	 count = tsv[j];
      }
      prev[i] = count;
      active_left &= ~m;
   }
   record.Append(TSeries(tStart, dT, dvu));
}

//======================================  Reset the adc module
void
adcMod::reset(void) {
   tCurrent = Time(0);
   prev.resize(channels.size());
   reset_stride();
}

//======================================  Reset the adc module
void
adcMod::reset_stride(void) {
   record.Clear();
   alarm_flags = 0;
}

//======================================  Set the select (alarm) mask.
void
adcMod::selectMask(mask_type mask) {
   select_mask = mask;
}

//======================================  Build a channel name from the ifo
static string
chan_from_template(const string& chan_tmpl, const string& ifo, int dcu,
		   int unit, int ch) {
   string chan = chan_tmpl;
   size_t inx = chan.find('%');
   while (inx != string::npos) {
      if (chan[inx+1] == 'd') {        //  test for dcu escape
	 string dcuid = to_string(dcu);
	 chan.replace(inx, 2, dcuid);
	 inx += dcuid.size();
      } else if (chan[inx+1] == 'i') { //  test for ifo escape
	 chan.replace(inx, 2, ifo);
	 inx += ifo.size();
      } else if (chan[inx+1] == 'c') { //  test for channel escape
	 string chnid = to_string(ch);
	 chan.replace(inx, 2, chnid);
	 inx += chnid.size();
      } else if (chan[inx+1] == 'u') { //  test for unit escape
	 string uid = to_string(unit);
	 chan.replace(inx, 2, uid);
	 inx += uid.size();
      } else {
	 inx ++;
      }
      inx = chan.find('%', inx);
   }
   return chan;
}

//======================================  Set the module control parameters
void
adcMod::setup(DaccAPI& in) {
   mask_type active_left = active_mask;
   for (size_t i=0; active_left; i++) {
      mask_type m = 1 << i;
      while (channels.size() <= i) channels.push_back("");
      if (!(m & active_mask)) continue;
      if (channels[i].empty()) {
	 channels[i] = chan_from_template(chan_template, ifoid, dcuid, unitid, i);
      }
      while (signals.size() <= i) signals.push_back("");
      if (signals[i].empty()) {
	 signals[i] = chan_from_template(signal_template, ifoid, dcuid, unitid, i);
      }
      in.addChannel(channels[i]);
      active_left &= ~m;
   }
}

//======================================  Set the overflow channel name
void
adcMod::set_channel(size_t id, const std::string& chan) {
   if (channels.size() <= id) channels.resize(id+1);
   channels[id] = chan;
}

//======================================  Set the overflow channel name template
void
adcMod::set_channel_template(const std::string& temp) {
   chan_template = temp;
}

//======================================  Set the signal name template
void
adcMod::set_signal_template(const std::string& temp) {
   signal_template = temp;
}

//======================================  Set the debug verbosity level
void
adcMod::set_debug(int lvl) {
   debug = lvl;
}

//======================================  Write data for this module
void
adcMod::write(FrWriter& w, const Time& t0, Interval dt) {
   string channel = chan_from_template("%i:FEC-%d_MODULE_%u_OVERFLOW",
				       ifoid, dcuid, unitid, -1);
   w.addProcSeries(channel, record.extract(t0, dt));
   record.eraseStart(t0 + dt - record.getStartTime());
}
